@@ -13,16 +13,18 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import logging
-
 from oslo_config import cfg
 from oslo_db import exception as odb_exceptions
+from oslo_log import log as logging
 from oslo_utils import excutils
 import pecan
 from stevedore import driver as stevedore_driver
 from wsme import types as wtypes
 from wsmeext import pecan as wsme_pecan
 
+from octavia.api.drivers import data_models as driver_dm
+from octavia.api.drivers import driver_factory
+from octavia.api.drivers import utils as driver_utils
 from octavia.api.v2.controllers import base
 from octavia.api.v2.controllers import l7policy
 from octavia.api.v2.types import listener as listener_types
@@ -30,9 +32,9 @@ from octavia.common import constants
 from octavia.common import data_models
 from octavia.common import exceptions
 from octavia.common import stats
-from octavia.common.tls_utils import cert_parser
 from octavia.db import api as db_api
 from octavia.db import prepare as db_prepare
+from octavia.i18n import _
 
 
 CONF = cfg.CONF
@@ -44,32 +46,23 @@ class ListenersController(base.BaseController):
 
     def __init__(self):
         super(ListenersController, self).__init__()
-        self.handler = self.handler.listener
         self.cert_manager = stevedore_driver.DriverManager(
             namespace='octavia.cert_manager',
             name=CONF.certificates.cert_manager,
             invoke_on_load=True,
         ).driver
 
-    def _get_db_listener(self, session, id):
-        """Gets a listener object from the database."""
-        listener = super(ListenersController, self)._get_db_listener(
-            session, id)
-        load_balancer_id = listener.load_balancer_id
-        db_listener = self.repositories.listener.get(
-            session, load_balancer_id=load_balancer_id, id=id)
-        if not db_listener:
-            LOG.info("Listener %s not found.", id)
-            raise exceptions.NotFound(
-                resource=data_models.Listener._name(), id=id)
-        return db_listener
-
     @wsme_pecan.wsexpose(listener_types.ListenerRootResponse, wtypes.text,
                          [wtypes.text], ignore_extra_args=True)
     def get_one(self, id, fields=None):
         """Gets a single listener's details."""
         context = pecan.request.context.get('octavia_context')
-        db_listener = self._get_db_listener(context.session, id)
+        db_listener = self._get_db_listener(context.session, id,
+                                            show_deleted=False)
+
+        if not db_listener:
+            raise exceptions.NotFound(resource=data_models.Listener._name(),
+                                      id=id)
 
         self._auth_validate_action(context, db_listener.project_id,
                                    constants.RBAC_GET_ONE)
@@ -121,13 +114,18 @@ class ListenersController(base.BaseController):
                 raise exceptions.ImmutableObject(resource=db_lb._name(),
                                                  id=lb_id)
 
-    def _validate_pool(self, session, lb_id, pool_id):
+    def _validate_pool(self, session, lb_id, pool_id, listener_protocol):
         """Validate pool given exists on same load balancer as listener."""
         db_pool = self.repositories.pool.get(
             session, load_balancer_id=lb_id, id=pool_id)
         if not db_pool:
             raise exceptions.NotFound(
                 resource=data_models.Pool._name(), id=pool_id)
+        if (db_pool.protocol == constants.PROTOCOL_UDP and
+                db_pool.protocol != listener_protocol):
+            msg = _("Listeners of type %s can only have pools of "
+                    "type UDP.") % constants.PROTOCOL_UDP
+            raise exceptions.ValidationException(detail=msg)
 
     def _reset_lb_status(self, session, lb_id):
         # Setting LB back to active because this should be a recoverable error
@@ -140,8 +138,8 @@ class ListenersController(base.BaseController):
         bad_refs = []
         for ref in tls_refs:
             try:
-                cert_parser.load_certificate_data(
-                    self.cert_manager, ref, context)
+                self.cert_manager.set_acls(context, ref)
+                self.cert_manager.get_cert(context, ref, check_only=True)
             except Exception:
                 bad_refs.append(ref)
 
@@ -191,24 +189,9 @@ class ListenersController(base.BaseController):
             raise exceptions.InvalidOption(value=listener_dict.get('protocol'),
                                            option='protocol')
 
-    def _send_listener_to_handler(self, session, db_listener):
-        try:
-            LOG.info("Sending Creation of Listener %s to handler",
-                     db_listener.id)
-            self.handler.create(db_listener)
-        except Exception:
-            with excutils.save_and_reraise_exception(
-                    reraise=False), db_api.get_lock_session() as lock_session:
-                self._reset_lb_status(
-                    lock_session, lb_id=db_listener.load_balancer_id)
-                # Listener now goes to ERROR
-                self.repositories.listener.update(
-                    lock_session, db_listener.id,
-                    provisioning_status=constants.ERROR)
-        db_listener = self._get_db_listener(session, db_listener.id)
-        result = self._convert_db_to_type(db_listener,
-                                          listener_types.ListenerResponse)
-        return listener_types.ListenerRootResponse(listener=result)
+    def _is_tls_or_insert_header(self, listener):
+        return (listener.default_tls_container_ref or
+                listener.sni_container_refs or listener.insert_headers)
 
     @wsme_pecan.wsexpose(listener_types.ListenerRootResponse,
                          body=listener_types.ListenerRootPOST, status_code=201)
@@ -218,17 +201,24 @@ class ListenersController(base.BaseController):
         context = pecan.request.context.get('octavia_context')
 
         load_balancer_id = listener.loadbalancer_id
-        listener.project_id = self._get_lb_project_id(
+        listener.project_id, provider = self._get_lb_project_id_provider(
             context.session, load_balancer_id)
 
         self._auth_validate_action(context, listener.project_id,
                                    constants.RBAC_POST)
-
+        if (listener.protocol == constants.PROTOCOL_UDP and
+                self._is_tls_or_insert_header(listener)):
+            raise exceptions.ValidationException(detail=_(
+                "%s protocol listener does not support TLS or header "
+                "insertion.") % constants.PROTOCOL_UDP)
         if (not CONF.api_settings.allow_tls_terminated_listeners and
                 listener.protocol == constants.PROTOCOL_TERMINATED_HTTPS):
             raise exceptions.DisabledOption(
                 value=constants.PROTOCOL_TERMINATED_HTTPS, option='protocol')
 
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(provider)
+
         lock_session = db_api.get_session(autocommit=False)
         try:
             if self.repositories.check_quota_met(
@@ -244,19 +234,38 @@ class ListenersController(base.BaseController):
 
             if listener_dict['default_pool_id']:
                 self._validate_pool(context.session, load_balancer_id,
-                                    listener_dict['default_pool_id'])
+                                    listener_dict['default_pool_id'],
+                                    listener.protocol)
 
             self._test_lb_and_listener_statuses(
                 lock_session, lb_id=load_balancer_id)
 
             db_listener = self._validate_create_listener(
                 lock_session, listener_dict)
+
+            # Prepare the data for the driver data model
+            provider_listener = (
+                driver_utils.db_listener_to_provider_listener(db_listener))
+
+            # re-inject the sni container references lost due to SNI
+            # being a separate table in the DB
+            provider_listener.sni_container_refs = listener.sni_container_refs
+
+            # Dispatch to the driver
+            LOG.info("Sending create Listener %s to provider %s",
+                     db_listener.id, driver.name)
+            driver_utils.call_provider(
+                driver.name, driver.listener_create, provider_listener)
+
             lock_session.commit()
         except Exception:
             with excutils.save_and_reraise_exception():
                 lock_session.rollback()
 
-        return self._send_listener_to_handler(context.session, db_listener)
+        db_listener = self._get_db_listener(context.session, db_listener.id)
+        result = self._convert_db_to_type(db_listener,
+                                          listener_types.ListenerResponse)
+        return listener_types.ListenerRootResponse(listener=result)
 
     def _graph_create(self, lock_session, listener_dict,
                       l7policies=None, pool_name_ids=None):
@@ -266,7 +275,8 @@ class ListenersController(base.BaseController):
         l7policies = listener_dict.pop('l7policies', l7policies)
         if listener_dict.get('default_pool_id'):
             self._validate_pool(lock_session, load_balancer_id,
-                                listener_dict['default_pool_id'])
+                                listener_dict['default_pool_id'],
+                                listener_dict['protocol'])
         db_listener = self._validate_create_listener(
             lock_session, listener_dict)
 
@@ -286,21 +296,8 @@ class ListenersController(base.BaseController):
                 l7p['redirect_pool_id'] = pool_id
             new_l7ps.append(l7policy.L7PolicyController()._graph_create(
                 lock_session, l7p))
-        return db_listener, new_l7ps
-
-    def _set_default_on_none(self, listener):
-        """Reset settings to their default values if None/null was passed in
-
-        A None/null value can be passed in to clear a value. PUT values
-        that were not provided by the user have a type of wtypes.UnsetType.
-        If the user is attempting to clear values, they should either
-        be set to None (for example in the name field) or they should be
-        reset to their default values.
-        This method is intended to handle those values that need to be set
-        back to a default value.
-        """
-        if listener.connection_limit is None:
-            listener.connection_limit = constants.DEFAULT_CONNECTION_LIMIT
+        db_listener.l7policies = new_l7ps
+        return db_listener
 
     @wsme_pecan.wsexpose(listener_types.ListenerRootResponse, wtypes.text,
                          body=listener_types.ListenerRootPUT, status_code=200)
@@ -308,11 +305,14 @@ class ListenersController(base.BaseController):
         """Updates a listener on a load balancer."""
         listener = listener_.listener
         context = pecan.request.context.get('octavia_context')
-        db_listener = self._get_db_listener(context.session, id)
+        db_listener = self._get_db_listener(context.session, id,
+                                            show_deleted=False)
         load_balancer_id = db_listener.load_balancer_id
 
-        self._auth_validate_action(context, db_listener.project_id,
-                                   constants.RBAC_PUT)
+        project_id, provider = self._get_lb_project_id_provider(
+            context.session, load_balancer_id)
+
+        self._auth_validate_action(context, project_id, constants.RBAC_PUT)
 
         # TODO(rm_work): Do we need something like this? What do we do on an
         # empty body for a PUT?
@@ -320,13 +320,15 @@ class ListenersController(base.BaseController):
             raise exceptions.ValidationException(
                 detail='No listener object supplied.')
 
-        self._set_default_on_none(listener)
+        if (db_listener.protocol == constants.PROTOCOL_UDP and
+                self._is_tls_or_insert_header(listener)):
+            raise exceptions.ValidationException(detail=_(
+                "%s protocol listener does not support TLS or header "
+                "insertion.") % constants.PROTOCOL_UDP)
 
         if listener.default_pool_id:
             self._validate_pool(context.session, load_balancer_id,
-                                listener.default_pool_id)
-        self._test_lb_and_listener_statuses(context.session, load_balancer_id,
-                                            id=id)
+                                listener.default_pool_id, db_listener.protocol)
 
         sni_containers = listener.sni_container_refs or []
         tls_refs = [sni for sni in sni_containers]
@@ -334,18 +336,38 @@ class ListenersController(base.BaseController):
             tls_refs.append(listener.default_tls_container_ref)
         self._validate_tls_refs(tls_refs)
 
-        try:
-            LOG.info("Sending Update of Listener %s to handler", id)
-            self.handler.update(db_listener, listener)
-        except Exception:
-            with excutils.save_and_reraise_exception(
-                    reraise=False), db_api.get_lock_session() as lock_session:
-                self._reset_lb_status(
-                    lock_session, lb_id=db_listener.load_balancer_id)
-                # Listener now goes to ERROR
-                self.repositories.listener.update(
-                    lock_session, db_listener.id,
-                    provisioning_status=constants.ERROR)
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(provider)
+
+        with db_api.get_lock_session() as lock_session:
+            self._test_lb_and_listener_statuses(lock_session,
+                                                load_balancer_id, id=id)
+
+            # Prepare the data for the driver data model
+            listener_dict = listener.to_dict(render_unsets=False)
+            listener_dict['id'] = id
+            provider_listener_dict = (
+                driver_utils.listener_dict_to_provider_dict(listener_dict))
+
+            # Also prepare the baseline object data
+            old_provider_llistener = (
+                driver_utils.db_listener_to_provider_listener(db_listener))
+
+            # Dispatch to the driver
+            LOG.info("Sending update Listener %s to provider %s", id,
+                     driver.name)
+            driver_utils.call_provider(
+                driver.name, driver.listener_update,
+                old_provider_llistener,
+                driver_dm.Listener.from_dict(provider_listener_dict))
+
+            # Update the database to reflect what the driver just accepted
+            self.repositories.listener.update(
+                lock_session, id, **listener.to_dict(render_unsets=False))
+
+        # Force SQL alchemy to query the DB, otherwise we get inconsistent
+        # results
+        context.session.expire_all()
         db_listener = self._get_db_listener(context.session, id)
         result = self._convert_db_to_type(db_listener,
                                           listener_types.ListenerResponse)
@@ -355,32 +377,69 @@ class ListenersController(base.BaseController):
     def delete(self, id):
         """Deletes a listener from a load balancer."""
         context = pecan.request.context.get('octavia_context')
-        db_listener = self._get_db_listener(context.session, id)
+        db_listener = self._get_db_listener(context.session, id,
+                                            show_deleted=False)
         load_balancer_id = db_listener.load_balancer_id
 
-        self._auth_validate_action(context, db_listener.project_id,
-                                   constants.RBAC_DELETE)
+        project_id, provider = self._get_lb_project_id_provider(
+            context.session, load_balancer_id)
 
-        if db_listener.provisioning_status == constants.DELETED:
-            return
+        self._auth_validate_action(context, project_id, constants.RBAC_DELETE)
 
-        self._test_lb_and_listener_statuses(
-            context.session, load_balancer_id,
-            id=id, listener_status=constants.PENDING_DELETE)
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(provider)
 
-        try:
-            LOG.info("Sending Deletion of Listener %s to handler",
-                     db_listener.id)
-            self.handler.delete(db_listener)
-        except Exception:
-            with excutils.save_and_reraise_exception(
-                    reraise=False), db_api.get_lock_session() as lock_session:
-                self._reset_lb_status(
-                    lock_session, lb_id=db_listener.load_balancer_id)
-                # Listener now goes to ERROR
-                self.repositories.listener.update(
-                    lock_session, db_listener.id,
-                    provisioning_status=constants.ERROR)
+        with db_api.get_lock_session() as lock_session:
+
+            self._test_lb_and_listener_statuses(
+                lock_session, load_balancer_id,
+                id=id, listener_status=constants.PENDING_DELETE)
+
+            LOG.info("Sending delete Listener %s to provider %s", id,
+                     driver.name)
+            provider_listener = (
+                driver_utils.db_listener_to_provider_listener(db_listener))
+            driver_utils.call_provider(driver.name, driver.listener_delete,
+                                       provider_listener)
+
+        # Revoke access of octavia service user to certificates
+        tls_refs = []
+
+        for sni in db_listener.sni_containers:
+            filters = {'tls_container_id': sni.tls_container_id}
+            snis = self.repositories.sni.get_all(context.session, **filters)[0]
+
+            if len(snis) == 1:
+                # referred only once, enqueue for access revoking
+                tls_refs.append(sni.tls_container_id)
+            else:
+                blocking_listeners = [s.listener_id for s in snis if
+                                      s.listener_id != id]
+                LOG.debug("Listeners %s using TLS ref %s. Access to TLS ref "
+                          "will not be revoked.", blocking_listeners,
+                          sni.tls_container_id)
+
+        if db_listener.tls_certificate_id:
+            filters = {'tls_certificate_id': db_listener.tls_certificate_id}
+            # Note get_all returns the list and links. We only want the list.
+            listeners = self.repositories.listener.get_all(
+                context.session, show_deleted=False, **filters)[0]
+
+            if len(listeners) == 1:
+                # referred only once, enqueue for access revoking
+                tls_refs.append(db_listener.tls_certificate_id)
+            else:
+                blocking_listeners = [l.id for l in listeners if l.id != id]
+                LOG.debug("Listeners %s using TLS ref %s. Access to TLS ref "
+                          "will not be revoked.", blocking_listeners,
+                          db_listener.tls_certificate_id)
+
+        for ref in tls_refs:
+            try:
+                self.cert_manager.unset_acls(context, ref)
+            except Exception:
+                # certificate may have been removed already
+                pass
 
     @pecan.expose()
     def _lookup(self, id, *remainder):
@@ -389,8 +448,9 @@ class ListenersController(base.BaseController):
         Currently it checks if this was a stats request and routes
         the request to the StatsController.
         """
-        if id and len(remainder) and remainder[0] == 'stats':
+        if id and remainder and remainder[0] == 'stats':
             return StatisticsController(listener_id=id), remainder[1:]
+        return None
 
 
 class StatisticsController(base.BaseController, stats.StatsMixin):
@@ -404,7 +464,8 @@ class StatisticsController(base.BaseController, stats.StatsMixin):
                          status_code=200)
     def get(self):
         context = pecan.request.context.get('octavia_context')
-        db_listener = self._get_db_listener(context.session, self.id)
+        db_listener = self._get_db_listener(context.session, self.id,
+                                            show_deleted=False)
         if not db_listener:
             LOG.info("Listener %s not found.", id)
             raise exceptions.NotFound(
