@@ -22,6 +22,9 @@ import pecan
 from wsme import types as wtypes
 from wsmeext import pecan as wsme_pecan
 
+from octavia.api.drivers import data_models as driver_dm
+from octavia.api.drivers import driver_factory
+from octavia.api.drivers import utils as driver_utils
 from octavia.api.v2.controllers import base
 from octavia.api.v2.controllers import listener
 from octavia.api.v2.controllers import pool
@@ -47,14 +50,19 @@ class LoadBalancersController(base.BaseController):
 
     def __init__(self):
         super(LoadBalancersController, self).__init__()
-        self.handler = self.handler.load_balancer
 
     @wsme_pecan.wsexpose(lb_types.LoadBalancerRootResponse, wtypes.text,
                          [wtypes.text], ignore_extra_args=True)
     def get_one(self, id, fields=None):
         """Gets a single load balancer's details."""
         context = pecan.request.context.get('octavia_context')
-        load_balancer = self._get_db_lb(context.session, id)
+        load_balancer = self._get_db_lb(context.session, id,
+                                        show_deleted=False)
+
+        if not load_balancer:
+            raise exceptions.NotFound(
+                resource=data_models.LoadBalancer._name(),
+                id=id)
 
         self._auth_validate_action(context, load_balancer.project_id,
                                    constants.RBAC_GET_ONE)
@@ -97,16 +105,6 @@ class LoadBalancersController(base.BaseController):
             raise exceptions.LBPendingStateError(
                 state=prov_status, id=id)
 
-    def _test_and_set_failover_prov_status(self, session, id):
-        lb_repo = self.repositories.load_balancer
-        if not lb_repo.set_status_for_failover(session, id,
-                                               constants.PENDING_UPDATE):
-            prov_status = lb_repo.get(session, id=id).provisioning_status
-            LOG.info("Invalid state %(state)s of loadbalancer resource %(id)s",
-                     {"state": prov_status, "id": id})
-            raise exceptions.LBPendingStateError(
-                state=prov_status, id=id)
-
     @staticmethod
     def _validate_network_and_fill_or_validate_subnet(load_balancer):
         network = validate.network_exists_optionally_contains_subnet(
@@ -250,15 +248,16 @@ class LoadBalancersController(base.BaseController):
 
         if not load_balancer.project_id:
             raise exceptions.ValidationException(detail=_(
-                "Missing project ID in request where one is required. "
-                "An administrator should check the keystone settings "
-                "in the Octavia configuration."))
+                "Missing project ID in request where one is required."))
 
         self._auth_validate_action(context, load_balancer.project_id,
                                    constants.RBAC_POST)
 
         self._validate_vip_request_object(load_balancer)
 
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(load_balancer.provider)
+
         lock_session = db_api.get_session(autocommit=False)
         try:
             if self.repositories.check_quota_met(
@@ -276,35 +275,61 @@ class LoadBalancersController(base.BaseController):
             ))
             vip_dict = lb_dict.pop('vip', {})
 
+            # Make sure we store the right provider in the DB
+            lb_dict['provider'] = driver.name
+
             # NoneType can be weird here, have to force type a second time
             listeners = lb_dict.pop('listeners', []) or []
             pools = lb_dict.pop('pools', []) or []
 
-            # TODO(johnsom) Remove provider and flavor from the lb_dict
-            # as they have not been implemented beyond the API yet.
-            # Remove these lines as they are implemented.
-            if 'provider' in lb_dict:
-                del lb_dict['provider']
-            if 'flavor_id' in lb_dict:
-                del lb_dict['flavor_id']
+            # TODO(johnsom) Remove flavor from the lb_dict
+            # as it has not been implemented beyond the API yet.
+            # Remove this line when it is implemented.
+            lb_dict.pop('flavor', None)
 
             db_lb = self.repositories.create_load_balancer_and_vip(
                 lock_session, lb_dict, vip_dict)
 
-            # create vip port if not exist
-            vip = self._create_vip_port_if_not_exist(db_lb)
+            # See if the provider driver wants to create the VIP port
+            octavia_owned = False
+            try:
+                provider_vip_dict = driver_utils.vip_dict_to_provider_dict(
+                    vip_dict)
+                vip_dict = driver_utils.call_provider(
+                    driver.name, driver.create_vip_port, db_lb.id,
+                    db_lb.project_id, provider_vip_dict)
+                vip = driver_utils.provider_vip_dict_to_vip_obj(vip_dict)
+            except exceptions.ProviderNotImplementedError:
+                # create vip port if not exist, driver didn't want to create
+                # the VIP port
+                vip = self._create_vip_port_if_not_exist(db_lb)
+                LOG.info('Created VIP port %s for provider %s.',
+                         vip.port_id, driver.name)
+                # If a port_id wasn't passed in and we made it this far
+                # we created the VIP
+                if 'port_id' not in vip_dict or not vip_dict['port_id']:
+                    octavia_owned = True
+
             self.repositories.vip.update(
-                lock_session, db_lb.id,
-                ip_address=vip.ip_address,
-                port_id=vip.port_id,
-                network_id=vip.network_id,
-                subnet_id=vip.subnet_id
-            )
+                lock_session, db_lb.id, ip_address=vip.ip_address,
+                port_id=vip.port_id, network_id=vip.network_id,
+                subnet_id=vip.subnet_id, octavia_owned=octavia_owned)
 
             if listeners or pools:
                 db_pools, db_lists = self._graph_create(
                     context.session, lock_session, db_lb, listeners, pools)
 
+            # Prepare the data for the driver data model
+            driver_lb_dict = driver_utils.lb_dict_to_provider_dict(
+                lb_dict, vip, db_pools, db_lists)
+
+            # Dispatch to the driver
+            LOG.info("Sending create Load Balancer %s to provider %s",
+                     db_lb.id, driver.name)
+            driver_utils.call_provider(
+                driver.name, driver.loadbalancer_create,
+                driver_dm.LoadBalancer.from_dict(driver_lb_dict))
+
             lock_session.commit()
         except odb_exceptions.DBDuplicateEntry:
             lock_session.rollback()
@@ -313,17 +338,6 @@ class LoadBalancersController(base.BaseController):
             with excutils.save_and_reraise_exception():
                 lock_session.rollback()
 
-        # Handler will be responsible for sending to controller
-        try:
-            LOG.info("Sending created Load Balancer %s to the handler",
-                     db_lb.id)
-            self.handler.create(db_lb)
-        except Exception:
-            with excutils.save_and_reraise_exception(reraise=False):
-                self.repositories.load_balancer.update(
-                    context.session, db_lb.id,
-                    provisioning_status=constants.ERROR)
-
         db_lb = self._get_db_lb(context.session, db_lb.id)
 
         result = self._convert_db_to_type(
@@ -358,7 +372,7 @@ class LoadBalancersController(base.BaseController):
                 pool_name = (
                     redirect_pool.get('name') if redirect_pool else None)
                 # All pools need to have a name so they can be referenced
-                if redirect_pool and not pool_name:
+                if default_pool and not pool_name:
                     raise exceptions.ValidationException(
                         detail='Pools must be named when creating a fully '
                                'populated loadbalancer.')
@@ -369,7 +383,7 @@ class LoadBalancersController(base.BaseController):
                     policy['redirect_pool'] = {'name': pool_name}
                     pools.append(redirect_pool)
                 # Otherwise, it's a reference and we record it and move on
-                elif redirect_pool:
+                elif default_pool:
                     pools_required.add(pool_name)
 
         # Make sure all pool names are unique.
@@ -405,9 +419,8 @@ class LoadBalancersController(base.BaseController):
                                                           attr=attr))
             p['load_balancer_id'] = db_lb.id
             p['project_id'] = db_lb.project_id
-            new_pool, new_hm, new_members = (
-                pool.PoolsController()._graph_create(
-                    session, lock_session, p))
+            new_pool = (pool.PoolsController()._graph_create(
+                session, lock_session, p))
             new_pools.append(new_pool)
             pool_name_ids[new_pool.name] = new_pool.id
 
@@ -444,25 +457,54 @@ class LoadBalancersController(base.BaseController):
         """Updates a load balancer."""
         load_balancer = load_balancer.loadbalancer
         context = pecan.request.context.get('octavia_context')
-        db_lb = self._get_db_lb(context.session, id)
+        db_lb = self._get_db_lb(context.session, id, show_deleted=False)
 
         self._auth_validate_action(context, db_lb.project_id,
                                    constants.RBAC_PUT)
 
-        if not isinstance(load_balancer.vip_qos_policy_id, wtypes.UnsetType):
-            network_driver = utils.get_network_driver()
-            validate.qos_extension_enabled(network_driver)
-            if load_balancer.vip_qos_policy_id is not None:
-                if db_lb.vip.qos_policy_id != load_balancer.vip_qos_policy_id:
-                    validate.qos_policy_exists(load_balancer.vip_qos_policy_id)
-        self._test_lb_status(context.session, id)
-        try:
-            LOG.info("Sending updated Load Balancer %s to the handler", id)
-            self.handler.update(db_lb, load_balancer)
-        except Exception:
-            with excutils.save_and_reraise_exception(reraise=False):
-                self.repositories.load_balancer.update(
-                    context.session, id, provisioning_status=constants.ERROR)
+        if (load_balancer.vip_qos_policy_id and
+                not isinstance(load_balancer.vip_qos_policy_id,
+                               wtypes.UnsetType) and
+                db_lb.vip.qos_policy_id != load_balancer.vip_qos_policy_id):
+            validate.qos_policy_exists(load_balancer.vip_qos_policy_id)
+
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(db_lb.provider)
+
+        with db_api.get_lock_session() as lock_session:
+            self._test_lb_status(lock_session, id)
+
+            # Prepare the data for the driver data model
+            lb_dict = load_balancer.to_dict(render_unsets=False)
+            lb_dict['id'] = id
+            vip_dict = lb_dict.pop('vip', {})
+            lb_dict = driver_utils.lb_dict_to_provider_dict(lb_dict)
+            if 'qos_policy_id' in vip_dict:
+                lb_dict['vip_qos_policy_id'] = vip_dict['qos_policy_id']
+
+            # Also prepare the baseline object data
+            old_provider_lb = (
+                driver_utils.db_loadbalancer_to_provider_loadbalancer(db_lb))
+
+            # Dispatch to the driver
+            LOG.info("Sending update Load Balancer %s to provider "
+                     "%s", id, driver.name)
+            driver_utils.call_provider(
+                driver.name, driver.loadbalancer_update,
+                old_provider_lb,
+                driver_dm.LoadBalancer.from_dict(lb_dict))
+
+            db_lb_dict = load_balancer.to_dict(render_unsets=False)
+            if 'vip' in db_lb_dict:
+                db_vip_dict = db_lb_dict.pop('vip')
+                self.repositories.vip.update(lock_session, id, **db_vip_dict)
+            if db_lb_dict:
+                self.repositories.load_balancer.update(lock_session, id,
+                                                       **db_lb_dict)
+
+        # Force SQL alchemy to query the DB, otherwise we get inconsistent
+        # results
+        context.session.expire_all()
         db_lb = self._get_db_lb(context.session, id)
         result = self._convert_db_to_type(db_lb, lb_types.LoadBalancerResponse)
         return lb_types.LoadBalancerRootResponse(loadbalancer=result)
@@ -472,13 +514,13 @@ class LoadBalancersController(base.BaseController):
         """Deletes a load balancer."""
         context = pecan.request.context.get('octavia_context')
         cascade = strutils.bool_from_string(cascade)
-        db_lb = self._get_db_lb(context.session, id)
+        db_lb = self._get_db_lb(context.session, id, show_deleted=False)
 
         self._auth_validate_action(context, db_lb.project_id,
                                    constants.RBAC_DELETE)
 
-        if db_lb.provisioning_status == constants.DELETED:
-            return
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(db_lb.provider)
 
         with db_api.get_lock_session() as lock_session:
             if (db_lb.listeners or db_lb.pools) and not cascade:
@@ -489,14 +531,12 @@ class LoadBalancersController(base.BaseController):
             self._test_lb_status(lock_session, id,
                                  lb_status=constants.PENDING_DELETE)
 
-        try:
-            LOG.info("Sending deleted Load Balancer %s to the handler", id)
-            self.handler.delete(db_lb, cascade)
-        except Exception:
-            with excutils.save_and_reraise_exception(reraise=False):
-                self.repositories.load_balancer.update(
-                    context.session, id,
-                    provisioning_status=constants.ERROR)
+            LOG.info("Sending delete Load Balancer %s to provider %s",
+                     id, driver.name)
+            provider_loadbalancer = (
+                driver_utils.db_loadbalancer_to_provider_loadbalancer(db_lb))
+            driver_utils.call_provider(driver.name, driver.loadbalancer_delete,
+                                       provider_loadbalancer, cascade)
 
     @pecan.expose()
     def _lookup(self, id, *remainder):
@@ -508,10 +548,14 @@ class LoadBalancersController(base.BaseController):
         'statuses' is aliased here for backward compatibility with
         neutron-lbaas LBaaS v2 API.
         """
-        if id and len(remainder) and (remainder[0] == 'status' or
-                                      remainder[0] == 'statuses' or
-                                      remainder[0] == 'stats' or
-                                      remainder[0] == 'failover'):
+        is_children = (
+            id and remainder and (
+                remainder[0] == 'status' or remainder[0] == 'statuses' or (
+                    remainder[0] == 'stats' or remainder[0] == 'failover'
+                )
+            )
+        )
+        if is_children:
             controller = remainder[0]
             remainder = remainder[1:]
             if controller == 'status' or controller == 'statuses':
@@ -520,6 +564,7 @@ class LoadBalancersController(base.BaseController):
                 return StatisticsController(lb_id=id), remainder
             elif controller == 'failover':
                 return FailoverController(lb_id=id), remainder
+        return None
 
 
 class StatusController(base.BaseController):
@@ -533,7 +578,8 @@ class StatusController(base.BaseController):
                          status_code=200)
     def get(self):
         context = pecan.request.context.get('octavia_context')
-        load_balancer = self._get_db_lb(context.session, self.id)
+        load_balancer = self._get_db_lb(context.session, self.id,
+                                        show_deleted=False)
         if not load_balancer:
             LOG.info("Load balancer %s not found.", id)
             raise exceptions.NotFound(
@@ -560,7 +606,8 @@ class StatisticsController(base.BaseController, stats.StatsMixin):
                          status_code=200)
     def get(self):
         context = pecan.request.context.get('octavia_context')
-        load_balancer = self._get_db_lb(context.session, self.id)
+        load_balancer = self._get_db_lb(context.session, self.id,
+                                        show_deleted=False)
         if not load_balancer:
             LOG.info("Load balancer %s not found.", id)
             raise exceptions.NotFound(
@@ -587,18 +634,18 @@ class FailoverController(LoadBalancersController):
     def put(self, **kwargs):
         """Fails over a loadbalancer"""
         context = pecan.request.context.get('octavia_context')
-        db_lb = self._get_db_lb(context.session, self.lb_id)
+        db_lb = self._get_db_lb(context.session, self.lb_id,
+                                show_deleted=False)
 
         self._auth_validate_action(context, db_lb.project_id,
                                    constants.RBAC_PUT_FAILOVER)
 
-        self._test_and_set_failover_prov_status(context.session, self.lb_id)
-        try:
-            LOG.info("Sending failover request for lb %s to the handler",
-                     self.lb_id)
-            self.handler.failover(db_lb)
-        except Exception:
-            with excutils.save_and_reraise_exception(reraise=False):
-                self.repositories.load_balancer.update(
-                    context.session, self.lb_id,
-                    provisioning_status=constants.ERROR)
+        # Load the driver early as it also provides validation
+        driver = driver_factory.get_driver(db_lb.provider)
+
+        with db_api.get_lock_session() as lock_session:
+            self._test_lb_status(lock_session, self.lb_id)
+            LOG.info("Sending failover request for load balancer %s to the "
+                     "provider %s", self.lb_id, driver.name)
+            driver_utils.call_provider(
+                driver.name, driver.loadbalancer_failover, self.lb_id)
