@@ -102,9 +102,22 @@ class BaseRepository(object):
         :param filters: Filters to decide which entity should be retrieved.
         :returns: octavia.common.data_model
         """
-        model = session.query(self.model_class).filter_by(**filters).first()
+        deleted = filters.pop('show_deleted', True)
+        model = session.query(self.model_class).filter_by(**filters)
+
+        if not deleted:
+            if hasattr(self.model_class, 'status'):
+                model = model.filter(
+                    self.model_class.status != consts.DELETED)
+            else:
+                model = model.filter(
+                    self.model_class.provisioning_status != consts.DELETED)
+
+        model = model.first()
+
         if not model:
-            return
+            return None
+
         return model.to_data_model()
 
     def get_all(self, session, pagination_helper=None,
@@ -213,13 +226,6 @@ class Repositories(object):
             if listener_id:
                 self.listener.update(session, listener_id,
                                      default_pool_id=pool_dict['id'])
-
-        # Immediate refresh, as we have found that sqlalchemy will sometimes
-        # cache the above query and the pool object may miss the listener_id
-        # information
-        if listener_id:
-            pool = session.query(models.Pool).filter_by(id=db_pool.id).first()
-            session.refresh(pool)
         return self.pool.get(session, id=db_pool.id)
 
     def update_pool_and_sp(self, session, pool_id, pool_dict):
@@ -244,7 +250,7 @@ class Repositories(object):
                     self.session_persistence.create(session, **sp_dict)
             # If only the session_persistence is being updated, this will be
             # empty
-            if len(pool_dict.keys()) > 0:
+            if pool_dict:
                 self.pool.update(session, pool_id, **pool_dict)
         return self.pool.get(session, id=pool_id)
 
@@ -353,8 +359,7 @@ class Repositories(object):
                 if lb_count <= lb_quota or lb_quota == consts.QUOTA_UNLIMITED:
                     quotas.in_use_load_balancer = lb_count
                     return False
-                else:
-                    return True
+                return True
             elif _class == data_models.Listener:
                 # Decide which quota to use
                 if quotas.listener is None:
@@ -375,8 +380,7 @@ class Repositories(object):
                         listener_quota == consts.QUOTA_UNLIMITED):
                     quotas.in_use_listener = listener_count
                     return False
-                else:
-                    return True
+                return True
             elif _class == data_models.Pool:
                 # Decide which quota to use
                 if quotas.pool is None:
@@ -397,8 +401,7 @@ class Repositories(object):
                         pool_quota == consts.QUOTA_UNLIMITED):
                     quotas.in_use_pool = pool_count
                     return False
-                else:
-                    return True
+                return True
             elif _class == data_models.HealthMonitor:
                 # Decide which quota to use
                 if quotas.health_monitor is None:
@@ -419,8 +422,7 @@ class Repositories(object):
                         hm_quota == consts.QUOTA_UNLIMITED):
                     quotas.in_use_health_monitor = hm_count
                     return False
-                else:
-                    return True
+                return True
             elif _class == data_models.Member:
                 # Decide which quota to use
                 if quotas.member is None:
@@ -441,8 +443,7 @@ class Repositories(object):
                         member_quota == consts.QUOTA_UNLIMITED):
                     quotas.in_use_member = member_count
                     return False
-                else:
-                    return True
+                return True
         except db_exception.DBDeadlock:
             LOG.warning('Quota project lock timed out for project: %(proj)s',
                         {'proj': project_id})
@@ -775,9 +776,8 @@ class LoadBalancerRepository(BaseRepository):
             # If a load balancer was never updated use its creation timestamp
             last_update = lb.updated_at or lb.created_at
             return last_update < timestamp
-        else:
-            # Load balancer was just deleted.
-            return True
+        # Load balancer was just deleted.
+        return True
 
 
 class VipRepository(BaseRepository):
@@ -950,8 +950,7 @@ class ListenerRepository(BaseRepository):
                 max_peer_port = listener.peer_port
         if max_peer_port == 0:
             return consts.HAPROXY_BASE_PEER_PORT
-        else:
-            return max_peer_port + 1
+        return max_peer_port + 1
 
     def _pool_check(self, session, pool_id, listener_id=None,
                     lb_id=None):
@@ -1074,7 +1073,6 @@ class AmphoraRepository(BaseRepository):
                 id=amphora_id).first()
             load_balancer.amphorae.append(amphora)
 
-    @oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
     def allocate_and_associate(self, session, load_balancer_id):
         """Allocate an amphora for a load balancer.
 
@@ -1125,6 +1123,34 @@ class AmphoraRepository(BaseRepository):
             ).first()
             if db_lb:
                 return db_lb.to_data_model()
+            return None
+
+    def get_all_deleted_expiring_amphora(self, session, exp_age=None):
+
+        """Get all previously deleted amphora that are now expiring.
+
+        :param session: A Sql Alchemy database session.
+        :param exp_age: A standard datetime delta which is used to see for how
+                        long can an amphora live without updates before it is
+                        considered expired (default:
+                        CONF.house_keeping.amphora_expiry_age)
+        :returns: [octavia.common.data_model]
+        """
+        if not exp_age:
+            exp_age = datetime.timedelta(
+                seconds=CONF.house_keeping.amphora_expiry_age)
+
+        expiry_time = datetime.datetime.utcnow() - exp_age
+
+        query = session.query(self.model_class).filter_by(
+            status=consts.DELETED).filter(
+            self.model_class.updated_at < expiry_time)
+        # Only make one trip to the database
+        query = query.options(joinedload('*'))
+        model_list = query.all()
+
+        data_model_list = [model.to_data_model() for model in model_list]
+        return data_model_list
 
     def get_spare_amphora_count(self, session):
         """Get the count of the spare amphora.
@@ -1340,8 +1366,8 @@ class AmphoraHealthRepository(BaseRepository):
                 model_kwargs['amphora_id'] = amphora_id
                 self.create(session, **model_kwargs)
 
-    def check_amphora_expired(self, session, amphora_id, exp_age=None):
-        """check if a specific amphora is expired
+    def check_amphora_health_expired(self, session, amphora_id, exp_age=None):
+        """check if a specific amphora is expired in the amphora_health table
 
         :param session: A Sql Alchemy database session.
         :param amphora_id: id of an amphora object
@@ -1355,16 +1381,28 @@ class AmphoraHealthRepository(BaseRepository):
             exp_age = datetime.timedelta(
                 seconds=CONF.house_keeping.amphora_expiry_age)
 
-        timestamp = datetime.datetime.utcnow() - exp_age
-        amphora_health = self.get(session, amphora_id=amphora_id)
-        if amphora_health is not None:
-            return amphora_health.last_update < timestamp
-        else:
-            # Amphora was just destroyed.
-            return True
+        expiry_time = datetime.datetime.utcnow() - exp_age
+
+        amphora_model = (
+            session.query(models.AmphoraHealth)
+            .filter_by(amphora_id=amphora_id)
+            .filter(models.AmphoraHealth.last_update > expiry_time)
+        ).first()
+        # This will return a value if:
+        # * there is an entry in the table for this amphora_id
+        # AND
+        # * the entry was last updated more recently than our expiry_time
+        # Receiving any value means that the amp is unexpired.
+
+        # In contrast, we receive no value if:
+        # * there is no entry for this amphora_id
+        # OR
+        # * the entry was last updated before our expiry_time
+        # In this case, the amphora is expired.
+        return amphora_model is None
 
     def get_stale_amphora(self, session):
-        """Retrieves a staled amphora from the health manager database.
+        """Retrieves a stale amphora from the health manager database.
 
         :param session: A Sql Alchemy database session.
         :returns: [octavia.common.data_model]
@@ -1452,10 +1490,6 @@ class L7RuleRepository(BaseRepository):
         with session.begin(subtransactions=True):
             if not model_kwargs.get('id'):
                 model_kwargs.update(id=uuidutils.generate_uuid())
-            if model_kwargs.get('l7policy_id'):
-                l7policy_db = session.query(models.L7Policy).filter_by(
-                    id=model_kwargs.get('l7policy_id')).first()
-                model_kwargs.update(l7policy=l7policy_db)
             l7rule = self.model_class(**model_kwargs)
             validate.l7rule_data(l7rule)
             session.add(l7rule)
@@ -1594,10 +1628,6 @@ class L7PolicyRepository(BaseRepository):
                 pool_db = session.query(models.Pool).filter_by(
                     id=model_kwargs.get('redirect_pool_id')).first()
                 model_kwargs.update(redirect_pool=pool_db)
-            if model_kwargs.get('listener_id'):
-                listener_db = session.query(models.Listener).filter_by(
-                    id=model_kwargs.get('listener_id')).first()
-                model_kwargs.update(listener=listener_db)
             l7policy = self.model_class(
                 **validate.sanitize_l7policy_api_args(model_kwargs,
                                                       create=True))
